#include <assert.h>
#include <qobject.h>
#include <qdatastream.h>
#include <math.h>

#include "cue.h"
#include "slight.h"
#include "slmodel.h"

Cue::Cue( float fadeTime ) 
    : _name(QString::null), _num(0), _fadeTime( fadeTime ), 
    _follow(false), _followTime(0), _jumpToCue(NULL), _jumpTimes(0)
{
}

Cue::Cue( Cue& p_other )
    : _levels( p_other._levels ), _name( p_other._name ), _num( p_other._num ),
    _fadeTime( p_other._fadeTime ), _follow( p_other._follow ),
    _followTime( p_other._followTime ), _jumpToCue( p_other._jumpToCue ),
    _jumpTimes( p_other._jumpTimes )
{
}

Cue::~Cue() 
{
}

bool Cue::serialize( QDataStream& p_os )
{
    p_os << _name;
    p_os << _num;
    p_os << _fadeTime;
    if( _follow )
    {
        p_os << (unsigned int)1;
        p_os << _followTime;
    }
    else
        p_os << (unsigned int)0;
    
    // TODO: ditch this dependence on the model.
    // (eg, make the jumpToCue() a number or something, not a 
    // pointer that can't be serialized.)
    if( jumpToCueIsSet() )
        p_os << (model->findCue( _jumpToCue ) - model->findCue(this) );
    else
        p_os << (unsigned int)0;

    p_os << _jumpTimes;

    p_os << numChannels();

    begin();
    while( hasNext() )
    {
        p_os << currChannel() << currLevel();
        next();
    }

    return true;
}

bool Cue::restore( QDataStream& p_is )
{
    p_is >> _name;
    p_is >> _num;
    p_is >> _fadeTime;
    unsigned int follow = 0;
    p_is >> follow;
    if( follow == 1 )
    {
        _follow = true;
        p_is >> _followTime;
    }

    unsigned int jumpOffset;
    p_is >> jumpOffset;

    _jumpToCue = (Cue*)jumpOffset;

    p_is >> _jumpTimes;

    unsigned int numLevels;
    p_is >> numLevels;

    unsigned int maxchan = 0;
    for( unsigned int i = 0; i < numLevels; i++ )
    {
        unsigned int chan;
        int level;

        p_is >> chan >> level;
        
        setChannel( chan, level );
	if(chan > maxchan)
		maxchan = chan;

    }

    if(maxchan > model->getNumFaders()) {
	    printf("cue->restore: expanding to %d channels\n", maxchan+1);
	    model->setNumFaders(maxchan+1);
    }

    return true;
}

static void
fmtf(FILE *fp, char *s, double d)
{
	double ipart, fpart;

	fpart = modf(d, &ipart);

	if(fabs(d-ipart) < 0.0001) {
		fprintf(fp, "%s%.f", s, d);
	} else {
		fprintf(fp, "%s%.2f", s, d);
	}
}


void Cue::toascii( FILE *fp, bool percent )
{
	struct {int chan; int val;} cchans[20];

	fmtf(fp, "\nCue ", _num);
	fmtf(fp, "  Fade ", _fadeTime);
	fprintf(fp, "  \"%s\"", (const char *)_name);

	if( _follow ) {
		fmtf(fp, "  Follow ", _followTime);
	}
    
	if( jumpToCueIsSet() ) {
		fmtf(fp, "  Jump to ", _jumpToCue->number());
		fprintf(fp,  " %d times", _jumpTimes);
	}
	
	int rchan = 0; // number of channels printable so far in current row
	begin();
	while( hasNext() )
	{
		cchans[rchan].chan = currChannel()+1;
		if(percent)
			cchans[rchan].val =  (int)((float)currLevel() / 2.55);
		else
			cchans[rchan].val = currLevel();
		
		if(++rchan >= 20) {
			int i;
			fprintf(fp, "\n");
			for(i = 0; i < rchan; i++) {
				fprintf(fp, "%3d ", cchans[i].chan);
			}
			fprintf(fp, "\n");
			for(i = 0; i < rchan; i++) {
				fprintf(fp, "%3d ", cchans[i].val);
			}
			fprintf(fp, "\n");
			rchan = 0;
		}
		next();
	}
	if(rchan) {
		int i;
		fprintf(fp, "\n");
		for(i = 0; i < rchan; i++) {
			fprintf(fp, "%3d ", cchans[i].chan);
		}
		fprintf(fp, "\n");
		for(i = 0; i < rchan; i++) {
			fprintf(fp, "%3d ", cchans[i].val);
		}
		fprintf(fp, "\n");
		rchan = 0;
	}
}

void Cue::setChannel( unsigned int channel, int level )
{
    _levels[channel] = level;
}

void Cue::clearChannel( unsigned int channel )
{
    _levels.remove( channel );  
}

bool Cue::channelIsSet( unsigned int channel )
{
    return _levels.contains( channel );
}

int Cue::getChannel( unsigned int channel )
{
    if( _levels.contains( channel ) )
        return _levels[channel];
    else
        return -1;
}


unsigned int Cue::numChannels()
{
    return _levels.count();
}        

void Cue::begin()
{
    _it = _levels.begin();
}

void Cue::next()
{
    _it++;
}

bool Cue::hasNext()
{
    return _it != _levels.end();
}

unsigned int Cue::currChannel()
{
    return _it.key();
}

int Cue::currLevel()
{
    return _it.data();
}

void Cue::setFadeTime( float fadeTime )
{
    _fadeTime = fadeTime;
}

float Cue::fadeTime()
{
    return _fadeTime;
}

QString Cue::name()
{
    return _name;
}

void Cue::setName( QString& name )
{
    _name = name;
}

float Cue::number()
{
    return _num;
}

void Cue::setNumber( float num )
{
    _num = num;
}

float Cue::followTime()
{
    return _followTime;
}

void Cue::setFollowTime( float f )
{
    _followTime = f;
}

void Cue::setFollow( bool b )
{
    _follow = b;
}

bool Cue::followIsSet()
{
    return _follow;
}

Cue* Cue::jumpToCue()
{
    return _jumpToCue;
}

void Cue::setJumpToCue( Cue* c )
{
    _jumpToCue = c;
}

bool Cue::jumpToCueIsSet()
{
    return (_jumpToCue != NULL);
}

int Cue::jumpTimes()
{
    return _jumpTimes;
}

void Cue::setJumpTimes( int times ) 
{
    _jumpTimes = times;
}

