#include <math.h>

#include <qapplication.h>
#include <qthread.h>
#include <qevent.h>

#include "rtcfader.h"
#include "rtcfaderthread.h"
#include "dmxengine.h"

class RTCFaderThread;

RTCFaderThread* RTCFader::m_thread = NULL;
DMXEngine* RTCFader::m_engine = NULL;
int RTCFader::m_timerFrequency = 0;

RTCFader::RTCFader( unsigned int faderNum, DMXEngine* engine )
  : m_faderNum( faderNum ), m_level( 0 ), m_targetLevel( 0 ), m_active(false)
{
    if( m_engine == NULL )
        m_engine = engine;

    if( m_thread == NULL )
        m_thread = new RTCFaderThread( m_engine );
}

RTCFader::~RTCFader()
{
}

int RTCFader::getLevel()
{
    return (int)m_level;
}

void RTCFader::setLevel( int p_level )
{
    m_level = p_level;
    m_active = false;
    m_targetLevel = p_level;

    for( QValueList<unsigned int>::iterator i = m_channels.begin();
            i != m_channels.end(); i++ )
        m_engine->setLevel( *i, (int)m_level );

    if( !m_thread->running() )
        m_engine->effectLevelChange();

    // else will get called on the next thread update, and removed from
    // the list of active faders

    QThread::postEvent( qApp->mainWidget(), 
                    new QCustomEvent( QEvent::User, (void*)m_faderNum ));
}



/*
 * integer-only fade computation by Steve Tell
 * This is a generalization of the Midpoint or Bresenham line drawing algorithm
 * from computer graphics.
 * Assuming 32-bit integers, this could run into trouble if your fades
 * are more than a few years long.
 */
void RTCFader::fadeTo( int p_level, float p_time )
{
	int m, b, a;
	int nsteps;
	
    if( p_level == m_level )
        return;
    m_targetLevel = p_level;

    nsteps = (int) (m_timerFrequency * p_time);
    if(nsteps < 1) {
	    setLevel(p_level);
	    return;
    }	    

    m = abs((m_targetLevel - m_level)/nsteps);
    b = -nsteps;
    a = (int)m_targetLevel - (int)m_level;

    if(a < 0) {	/* SOUTH */
	    m_d = 2*a - b*(2*m + 1);
	    m_delta_d_below = 2*(a - b*m);	/* less SE */
	    m_delta_y_below = -m;
	    m_delta_d_above = 2*(a - b*(m+1));	/* more SE */
	    m_delta_y_above = -(m+1);
    } else {				/* NORTH */
	    m_d = 2*a + b*(2*m + 1);
	    m_delta_d_below = 2*(a + b*(m+1));	/* more NE */
	    m_delta_y_below = m+1;
	    m_delta_d_above = 2*(a + b*m);	/* less NE */
	    m_delta_y_above = m;
    }

    m_thread->addFader( this );
}

void RTCFader::timer_tick()
{
    if( m_level == m_targetLevel )
    {
        m_active = false;
        return;
    }

    unsigned char oldLevel = m_level;

    if(m_d <= 0) {
	    m_d     += m_delta_d_above;
	    m_level += m_delta_y_above;		
    } else {
	    m_d     += m_delta_d_below;
	    m_level += m_delta_y_below;		
    }

    if( oldLevel != m_level )
    {
        for( QValueList<unsigned int>::iterator i = m_channels.begin();
                i != m_channels.end(); i++ )
            m_engine->setLevel( *i, (int)m_level ); 
    }

    return;
}

