#include <stdio.h>
#include <fcntl.h>

#include <linux/version.h>

// if we're running a 2.4 kernel, use linux/rtc.h.  else, use
// linux/mc146818rtc.h.  if someone can give me a more precise 2.3
// kernel version where we went to the new rtc, tell me and i'll update
// this in the next release

#if LINUX_VERSION_CODE < 132095
#include <linux/mc146818rtc.h>
#else
#include <linux/rtc.h>
#endif

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>

#include <iostream>
using namespace std;

#include <qthread.h>
#include <qptrlist.h>
#include <qapplication.h>

#include "rtcfader.h"
#include "rtcfaderthread.h"
#include "dmxengine.h"
#include "slight.h"

RTCFaderThread::RTCFaderThread( DMXEngine* p_engine )
    : QThread(), m_engine( p_engine )
{
    m_rtc = open( "/dev/rtc", O_RDONLY );

    if( m_rtc == -1 )
    {
        cerr << "Can't open the real-time clock." << endl;
        ::exit(1);
    }

    // set the timer rate for 64 Hz
    // TODO: maybe try for something higher if root?
    int retval = ioctl( m_rtc, RTC_IRQP_SET, SLIGHT_TIMER_FREQ );
    if( retval == -1 )
    {
        cerr << "Error setting RTC rate to 64 Hz." << endl;
        ::exit(1);
    }

    RTCFader::m_timerFrequency = SLIGHT_TIMER_FREQ;
}

RTCFaderThread::~RTCFaderThread()
{
    ::close(m_rtc);
}


void RTCFaderThread::addFader( RTCFader* p_fader )
{
    // can't enqueue a single fader more than once.  to enforce this, all
    // handling of RTCFader.m_active is inside the lock.
	bool wasactive;
	m_faderMutex.lock();
	wasactive = p_fader->m_active;
	if(!p_fader->m_active) {
		m_faders.enqueue( p_fader );
		p_fader->m_active = true;
	}
	m_faderMutex.unlock();

	if(wasactive) {
//	    printf("RTCFader(%u): already active m_level=%d target=%d\n",
//		   p_fader->m_faderNum, p_fader->m_level, p_fader->m_targetLevel);
//	    checkFader(p_fader);
		
	}
	if( !running() )
		start();
}

void RTCFaderThread::run()
{
    // first off, try to set this thread's priority as close to real-time
    // as possible.  if it fails, it's okay.... but if you're doing quick
    // fades, you should run SLight suid root!
	// (why don't we use SCHED_RR here? - sgt)

    int retval = setpriority( PRIO_PROCESS, 0, -20 );
    
    unsigned long data;
    
    // for iterating over the queue 
    // (because we don't want to allocate anything during the timer loop)
    RTCFader* head;
    RTCFader* currFader;

    unsigned int panelUpdateCounter = 0;

    retval = ioctl( m_rtc, RTC_PIE_ON, 0 );
    if( retval == -1 )
    {
        cerr << "Error enabling periodic interrupts." << endl;
        ::exit(1);
    }

    while( !m_faders.isEmpty() )
    {
        // blocking call to read.
        retval = read( m_rtc, &data, sizeof( unsigned long ) );

        // hopefully it read correctly.

        m_faderMutex.lock();

        head = m_faders.head();
        do
        {
            currFader = m_faders.dequeue();
            currFader->timer_tick();
            if( currFader->isActive() )
                m_faders.enqueue( currFader );
            else
                if( currFader == head )
                    head = m_faders.head();
        }
        while( !m_faders.isEmpty() && m_faders.head() != head );

        m_faderMutex.unlock();

        m_engine->effectLevelChange();

        panelUpdateCounter++;
        if( panelUpdateCounter == SLIGHT_PANEL_UPDATE_PERIOD )
        {
            panelUpdateCounter = 0;
            QThread::postEvent( qApp->mainWidget(), 
                    new QCustomEvent( QEvent::User,  (void*)UINT_MAX ));
        }
    }

    // is this ugly or what?  (-;

    QThread::postEvent( qApp->mainWidget(), 
                    new QCustomEvent( QEvent::User, (void*)UINT_MAX ));

    // the thread terminates when there are no more timers that need
    // to be updated.  clean up now.

    retval = ioctl( m_rtc, RTC_PIE_OFF, 0 );
    if( retval == -1 )
    {
        cerr << "Error disabling periodic interrupts." << endl;
        ::exit(1);
    }
//    write(1, "rtcfaderthread: exit\n", 21);
}


/* 
 * assert that a fader is in the queue - print error message if not.
 */
void RTCFaderThread::checkFader( RTCFader* p_fader )
{
	bool foundit = false;
	RTCFader* head;
	RTCFader* currFader;

        m_faderMutex.lock();

        head = m_faders.head();

	// surely there's a better way to iterate the list than this...
        do
        {
            currFader = m_faders.dequeue();
	    if(currFader == p_fader)
		    foundit = true;
            if( currFader->isActive() )
                m_faders.enqueue( currFader );
            else
                if( currFader == head )
                    head = m_faders.head();
        }
        while( !m_faders.isEmpty() && m_faders.head() != head );
	
        m_faderMutex.unlock();

	if(!foundit) {
		printf("RTCFaderThread::check(%u): should be on queue but is not!\n", p_fader->m_faderNum);
	}
}
