#include <qtimer.h>
#include <qvaluevector.h>
#include <qmessagebox.h>
#include <qapplication.h>
#include <qthread.h>
#include <qfile.h>
#include <qdatastream.h>

#include "slmodel.h"
#include "cue.h"
#include "zerocue.h"
#include "rtcfader.h"
#include "dmxengine.h"

SLModel::SLModel( DMXEngine* p_engine, QObject *parent, const char* name )
: QObject( parent, name ), m_engine( p_engine ), m_numChannels(0),
    m_currCue(NULL)
{
    connect( &m_followTimer, SIGNAL(timeout()),
            this, SLOT( fadeToNextCue() ) );
}

SLModel::~SLModel()
{
}

void SLModel::serialize( QDataStream& p_os )
{
    char* magic = "SLIGHT";
    p_os << magic;

    int maj_version = MAJ_VERSION; 
    int min_version = MIN_VERSION; 
    int rev = REV_VERSION;

    p_os << maj_version << min_version << rev;
    p_os << m_numChannels;
    p_os << m_faders.count();
    p_os << m_faderMap.count();

    for( QIntDictIterator<RTCFader> it( m_faderMap ); it.current(); ++it)
        p_os << it.currentKey() << it.current()->number();

    // serialize number of cues

    p_os << m_cues.count();

    // serialize settable parameters for zero cue

    p_os << ZeroCue::singleton()->fadeTime();
    p_os << ZeroCue::singleton()->followTime();
    if( ZeroCue::singleton()->jumpToCueIsSet() )
        p_os << model->findCue( ZeroCue::singleton()->jumpToCue() );
    else
    {
        int j = 0;
        p_os << j;
    }
        
    p_os << ZeroCue::singleton()->jumpTimes();

    for( unsigned int i = 1; i < m_cues.count(); i++ )
        m_cues.at(i)->serialize( p_os );
}

void SLModel::toascii(FILE *fp)
{
	fprintf(fp, "SLIGHT %d.%d.%d\n", 
		MAJ_VERSION, MIN_VERSION, REV_VERSION);
	fprintf(fp, " Faders: %d  Dimmers: %d   FaderMap: %d\n",
		m_faders.count(), m_numChannels, m_faderMap.count());

	unsigned i;
	int faderused;
	int ndim;
	for (i = 0; i < m_faders.count(); i++) {
		faderused = 0;
		ndim = 0;
		for( QIntDictIterator<RTCFader> it( m_faderMap ); it.current(); ++it) {
			if(it.current()->number() != i)
				continue;
			if(!faderused)
				fprintf(fp, " %3d:", i + 1);
			faderused = 1;
			fprintf(fp, " %3ld", it.currentKey() + 1); // fader
			if((++ndim % 16) == 0)
				fprintf(fp, "\n     ");
		}
		if(faderused)
			fprintf(fp, "\n");
	}

/*	for( QIntDictIterator<RTCFader> it( m_faderMap ); it.current(); ++it)
		fprintf(fp, " %ld: %d\n", 
			it.currentKey(), // dimmer
			it.current()->number()); // fader
*/

	fprintf(fp, "Number of cues in show: %d\n", m_cues.count());

	fprintf(fp, "\fCue 0  Fade %.f Blackout", ZeroCue::singleton()->fadeTime());
	fprintf(fp, "  Follow %.f", ZeroCue::singleton()->followTime());
	if( ZeroCue::singleton()->jumpToCueIsSet() ) {
		fprintf(fp, " Jump to %.f %d times",
			ZeroCue::singleton()->jumpToCue()->number(),
			ZeroCue::singleton()->jumpTimes());
	}
	fprintf(fp, "\n");
        
	fflush(fp);
	for( unsigned int i = 1; i < m_cues.count(); i++ ) {
		Cue *c = m_cues.at(i);
//		fprintf(fp, "");
		c->toascii( fp, true );
	}
}

bool SLModel::checkFile( QDataStream& p_is )
{
    char* magic = new char[7];
    p_is >> magic;

    if( strncmp( magic, "SLIGHT", 6 ) != 0 )
    {
        QMessageBox::warning( NULL, "SLight", 
                QString( "I don't recognize this file" ) );
        return false;
   }

    delete magic;

    int maj_version, min_version, rev; 
    p_is >> maj_version >> min_version >> rev;

    if( !(maj_version == 0 && min_version == 9) )
    {
        QMessageBox::warning( NULL, "SLight", 
                QString( "I can't read files of version %1.%2.%3" )
                .arg(maj_version).arg(min_version).arg(rev) );
        return false;
    }

    return true;
}

void SLModel::restore( QDataStream& p_is )
{
    p_is >> m_numChannels;

    int numFaders;
    p_is >> numFaders;

    setNumFaders(0);  //clear current faders, channels
    setNumFaders( numFaders );
    setNumChannels( m_numChannels );

    int numMappedChannels;
    p_is >> numMappedChannels;

    for( int i = 0; i < numMappedChannels; i++ )
    {
        int channel, fader;
        p_is >> channel >> fader;
        if( fader != -1 )
            mapChannelToFader( (unsigned int)channel, (unsigned int)fader );
    }

    unsigned int numCues;
    p_is >> numCues;

    // deal with zero cue info

    float zeroFadeTime;
    p_is >> zeroFadeTime;
    ZeroCue::singleton()->setFadeTime( zeroFadeTime );

    float zeroFollowTime;
    p_is >> zeroFollowTime;
    ZeroCue::singleton()->setFollowTime( zeroFollowTime );

    int zeroJumpOffset;
    p_is >> zeroJumpOffset;
    ZeroCue::singleton()->setJumpToCue( (Cue*)zeroJumpOffset );

    int zeroJumpTimes;
    p_is >> zeroJumpTimes;
    ZeroCue::singleton()->setJumpTimes(zeroJumpTimes);
    

    for( unsigned int i = 1; i < numCues; i++ )
    {
        Cue* tmpCue = new Cue();
        tmpCue->restore( p_is );
        appendCue( tmpCue );
    }

    // restore makes cue->jumpToCue() an *offset*; must change it
    // to an actual Cue*.

    for( unsigned int i = 0; i < getNumCues(); i++ )
    {
        Cue* currCue = getCue(i);
        if( currCue->jumpToCueIsSet() )
            currCue->setJumpToCue(
                    getCue((unsigned int)(findCue(currCue) + 
                           (int)currCue->jumpToCue() )));
    }
}

// checkpoint: write a savefile named with current date and time
void SLModel::checkpoint()
{
	time_t now;
	struct tm *nowtm;
	char fname[64];
	time(&now);
	nowtm = localtime(&now);
	strftime(fname, sizeof(fname), "~slight-%d%b%y-%H%M.sli", nowtm);

	QFile f( fname );
	if ( f.open(IO_WriteOnly) ) {   
		QDataStream s( &f );
		model->serialize( s );
		f.close();
		//printf("wrote %s\n", fname);
	}
}

void SLModel::clear()
{
    m_cues.remove( ZeroCue::singleton() );
    ZeroCue::singleton()->setFadeTime(5);
    ZeroCue::singleton()->setFollow(false);
    ZeroCue::singleton()->setJumpToCue(NULL);

    m_cues.setAutoDelete(true);
    m_cues.clear();
    m_cues.setAutoDelete(false);

    /**
    m_faders.setAutoDelete(true);
    m_faders.clear();
    m_faders.setAutoDelete(false);

    */

//    m_faderMap.clear();
    
/* causes crash, don't know why.  so leave the patch & channels.
   should be an option.

  setNumFaders(0);
    setNumFaders(INITIAL_NUM_CHANNELS);
    setNumChannels(INITIAL_NUM_CHANNELS);
    for( unsigned int i = 0; i < INITIAL_NUM_CHANNELS; i++ )
	    mapChannelToFader(i, i);
*/
    
    m_currCue = NULL;
    m_runningCues.clear();
    m_jumpCount.clear();

    appendCue( ZeroCue::singleton() );
    goTo( ZeroCue::singleton() );
}

unsigned int SLModel::getNumCues()
{
    return m_cues.count(); 
}

Cue* SLModel::getCue( unsigned int p_index )
{
    return m_cues.at(p_index);
}

void SLModel::appendCue( Cue* p_cue )
{
    m_cues.append( p_cue );
}

void SLModel::insertCue( Cue* p_cue, unsigned int p_index )
{
    m_cues.insert( p_index, p_cue );
}

void SLModel::insertCue( Cue* p_cue, Cue* p_afterCue )
{
    if( m_cues.findRef( p_afterCue ) != -1 )
        m_cues.insert( m_cues.findRef( p_afterCue ) + 1, p_cue );
}

void SLModel::removeCue( Cue* p_cue )
{
    // old comment TODO: Better cleanup (eg, check for jump-to cues, etc.)
    // This cleans up jump-to-cue targets
    for( Cue* currCue = m_cues.first(); currCue != NULL;
            currCue = m_cues.next() )
    {
        if( currCue->jumpToCue() == p_cue )
        {
            QMessageBox::warning( 0, "SLight", 
                    QString("The cue you are deleting was the ") 
                    + "\"jump to\" target of\n"
                    + "Cue " + QString::number( currCue->number() ) + ": "
                    + currCue->name() + "\n"
                    + "The above cue's \"jump to\" setting has been cleared." );

            currCue->setJumpToCue( NULL );
        }
    }

    m_cues.remove( p_cue );
}

int SLModel::findCue( Cue* p_cue )
{
    return m_cues.findRef( p_cue );
}

void SLModel::goTo( Cue* p_cue )
{  
    while( m_runningCues.count() != 0 )
        m_runningCues.removeFirst();

    m_followTimer.stop();

    m_currCue = p_cue;

    emit cueStatusChanged();

    QValueVector<int> levels( getNumFaders(), 0 );

    int lastFullCueIndex = findCue( p_cue );

    while( getCue(lastFullCueIndex)->numChannels() != getNumFaders() )
        lastFullCueIndex--;

    Cue* currCue = getCue( lastFullCueIndex );

    while( currCue != p_cue )
    {
        for( currCue->begin(); currCue->hasNext(); currCue->next() )
            levels[ currCue->currChannel() ] = currCue->currLevel();

        currCue = getCue( findCue(currCue) + 1 );
    }

    // add levels of current cue, too.

    for( currCue->begin(); currCue->hasNext(); currCue->next() )
        levels[ currCue->currChannel() ] = currCue->currLevel();
    
    for( unsigned int i = 0; i < getNumFaders(); i++ )
        m_faders.at(i)->setLevel( levels[i] );

    /*
    QThread::postEvent( qApp->mainWidget(), 
            new QCustomEvent( QEvent::User + 1 ) );
            */
}

void SLModel::fadeTo( Cue* p_cue )
{
    m_currCue = p_cue;
    m_runningCues.append( p_cue );
    QTimer::singleShot( (int)(p_cue->fadeTime() * 1000.0), this, 
            SLOT(dequeueRunningCue()) );

    emit cueStatusChanged();

    float fadeTime = p_cue->fadeTime();

    RTCFader *fader;
    for( p_cue->begin(); p_cue->hasNext(); p_cue->next() ) {
//	    printf("SLModel::fadeTo(%.f): chan=%d level=%d\n",
//		   p_cue->number(), p_cue->currChannel(), p_cue->currLevel());
	    fader = m_faders.at(p_cue->currChannel());
	    if(fader)
		    fader->fadeTo( p_cue->currLevel(), fadeTime ); 
	    else
		    printf("SLModel::fadeTo(%.f): bogus channel %d\n",
			   p_cue->number(), p_cue->currChannel());
    }
}

void SLModel::fadeToCue(Cue *cue)
{
    // does this interrupt a follow cue?
    if( m_followTimer.isActive() )
        m_followTimer.stop();

    // is this a follow-cue?
    if( cue->followIsSet() )
        m_followTimer.start( (int)(cue->followTime() * 1000.0), true );

    fadeTo( cue );
}

void SLModel::fadeToNextCue()
{
    // does this interrupt a follow cue?
    if( m_followTimer.isActive() )
        m_followTimer.stop();

    Cue* nextCue;

    //find the next cue: deal with jumping.

    if( m_currCue->jumpToCueIsSet() )
    {
        if( m_jumpCount.contains( m_currCue ) )
        {
            if( m_jumpCount[m_currCue] != 0 )
            {
                m_jumpCount[m_currCue]--;
                nextCue = m_currCue->jumpToCue();
            }
            else
            {
                m_jumpCount.remove( m_currCue );

                unsigned int nextCueIndex = findCue( m_currCue ) + 1;

                if( nextCueIndex < getNumCues() )
                    nextCue = getCue( nextCueIndex );
                else
                    return;
            }
        }
        else
        {
            m_jumpCount.insert( m_currCue, (m_currCue->jumpTimes() - 1) );
            nextCue = m_currCue->jumpToCue();
        }
    }
    else
    {
        unsigned int nextCueIndex = findCue( m_currCue ) + 1;

        if( nextCueIndex < getNumCues() )
            nextCue = getCue( nextCueIndex );
        else
            return;
    }  

    // is this a follow-cue?

    if( nextCue->followIsSet() )
        m_followTimer.start( (int)(nextCue->followTime() * 1000.0), true );


    fadeTo( nextCue );

}

bool SLModel::isCurrent( Cue* p_cue )
{
    return (p_cue == m_currCue);
}

Cue* SLModel::getCurrentCue()
{
    return m_currCue;
}


bool SLModel::isRunning( Cue* p_cue )
{
    return (m_runningCues.findRef( p_cue ) != -1);
}
    
int SLModel::getNumJumps( Cue* p_cue )
{
    if( p_cue->jumpToCueIsSet() )
        if( m_jumpCount.contains( p_cue ) ) 
            return m_jumpCount[p_cue];
        else
            return p_cue->jumpTimes();
    else
        return -1;
}

int SLModel::getFaderLevel( unsigned int p_fader )
{
    return m_faders.at(p_fader)->getLevel();
}

void SLModel::setFaderLevel( uint p_fader, int p_level )
{
    m_faders.at(p_fader)->setLevel( p_level );
}

void SLModel::mapChannelToFader( unsigned int p_channel, unsigned int p_fader )
{
    if( m_faderMap.find( p_channel ) == NULL )
    {
        m_faderMap.insert( p_channel, m_faders.at(p_fader) );
        m_faders.at(p_fader)->addChannel( p_channel );
    }
}

void SLModel::removeChannelMap( unsigned int p_channel, unsigned int p_fader )
{
    m_faderMap.remove( p_channel );
    m_faders.at(p_fader)->removeChannel( p_channel );
}

bool SLModel::getFaderForChannel( unsigned int p_channel, 
        unsigned int* p_fader )
{
    RTCFader* tmpFader = m_faderMap.find( p_channel ); 
    if( tmpFader != NULL )
    {
        *p_fader = m_faders.findRef( tmpFader );
        return true;
    }
    else
        return false;
}

unsigned int SLModel::getNumChannels()
{
    return m_numChannels;
}

void SLModel::setNumChannels( unsigned int p_channels )
{
    if( p_channels < m_numChannels )
        for( unsigned int chan = p_channels - 1; chan < m_numChannels; chan++ )
            for( unsigned int fader = 0; 
                    fader < getNumFaders();
                    fader++ )
                model->removeChannelMap( chan, fader );	

    m_numChannels = p_channels;
    m_engine->setNumChannels( m_numChannels );
}


unsigned int SLModel::getNumFaders()
{
    return m_faders.count();
}

void SLModel::setNumFaders( unsigned int p_numFaders )
{
    if( m_faders.count() == p_numFaders )
        return;

    while( m_faders.count() != p_numFaders )
        if( m_faders.count() < p_numFaders )
        {
            RTCFader* tmpFader = new RTCFader( m_faders.count(), m_engine );
            m_faders.append( tmpFader );
        }
        else
        {
            RTCFader* tmpFader = m_faders.last();
            m_faders.removeLast();

            for( unsigned int i = 0; i < getNumChannels(); i++ )
                if( m_faderMap.find(i) == tmpFader )
                    m_faderMap.take(i);
            
            delete tmpFader;

            unsigned int faderNum = m_faders.count();
            for( unsigned int cue = 0; cue < getNumCues(); cue++ )
               getCue(cue)->clearChannel( faderNum ); 
        }

    emit numFadersChanged( p_numFaders );

    return;
}

void SLModel::dequeueRunningCue()
{
    m_runningCues.removeFirst();
    emit cueStatusChanged();
}

void SLModel::reset()
{
    m_jumpCount.clear();
    goTo( ZeroCue::singleton() );
}


int SLModel::getFaderForCue(Cue *p_cue, uint p_fader)
{

    QValueVector<int> levels( getNumFaders(), 0 );

    int lastFullCueIndex = findCue( p_cue );

    while( getCue(lastFullCueIndex)->numChannels() != getNumFaders() )
        lastFullCueIndex--;

    Cue* currCue = getCue( lastFullCueIndex );

    while( currCue != p_cue )
    {
        for( currCue->begin(); currCue->hasNext(); currCue->next() )
            levels[ currCue->currChannel() ] = currCue->currLevel();

        currCue = getCue( findCue(currCue) + 1 );
    }

    // add levels of desired cue, too.
    for( currCue->begin(); currCue->hasNext(); currCue->next() )
        levels[ currCue->currChannel() ] = currCue->currLevel();
    
    return levels[p_fader];
}
