#include "channel.h"
#include <qpainter.h>
#include <qevent.h>
#include <qdragobject.h>
#include <qcursor.h>

#include <iostream>
using namespace std;


Channel::Channel( QWidget *parent, const char *name )
  : QWidget( parent, name ) 
{
  _level = 0;
  _levelDisplay = 0;
  _state = CS_NONE;
  _number = 0;
  setPalette( QPalette( QColor( 0, 0, 0) ) );

  _selected = false;
  _recorded = false;
  
  setAcceptDrops( true );
}

void Channel::setLevel( int level )
{
  if ( level == _level )
    return;
  _level = level;

  if( _percent )
    setLevelDisplay( (int)((float)_level / 2.55) );
  else
    setLevelDisplay( _level );

  emit levelChanged( _number, _level );
}

void Channel::setLevelDisplay( int level )
{
  _levelDisplay = level;
  update();
}

void Channel::confirmLevelDisplay()
{
  if( _percent )
    {
      _level = (int)((float)_levelDisplay * 2.55 );

      while( (int)((float)_level / 2.55 ) != _levelDisplay )
	_level++;

      _state = CS_CHANGED;
    }
  else
    _level = _levelDisplay;
  
  emit levelChanged( _number, _level );
}


void Channel::revertLevelDisplay()
{
  if( _percent )
    _levelDisplay = (int)((float)_level / 2.55 );
  else
    _levelDisplay = _level;
  
  update();
}

void Channel::setState( enum ChannelState state )
{
  if( state == _state )
    return;
  _state = state;

  update();

  emit stateChanged( _number, _state );
}

void Channel::select()
{
  if( _selected )
    return;

  _selected = true;
  update();
  emit selected();
  
  return;
}


void Channel::deselect()
{
  if( !_selected )
    return;

  _selected = false;
  update();
  emit deselected();
  return;
}


void Channel::paintEvent( QPaintEvent * )
{
   QPainter p( this );

   QColor chanColor = gray;
   QColor levelColor;
   
   switch( _state )
     {
     case CS_UP:
       levelColor = green;
       break;
       
     case CS_DOWN:
       levelColor = cyan;
       break;
       
     case CS_SAME:
       levelColor = magenta;
       break;

     case CS_CHANGED:
       levelColor = red;
       break;
       
     case CS_NONE:
       levelColor = gray;
       break;
       
     }
   
   if( _recorded )
     chanColor = white;
   
   // is this channel selected?  if so, draw the red box.
   if( _selected )
     {
       p.setBrush( red );
       p.setPen( NoPen );
       p.drawRoundRect ( 0, 0, width(), height(), 50, 50 );
       chanColor = yellow;
       levelColor = white;
     }

   p.setFont( QFont( "fixed", 12, QFont::Bold ) );
   p.setPen( QPen( chanColor ) );

   QRect text_rect = p.boundingRect( 0, 0, width(), height(), 0,  
				     QString::number( _number ) );

   p.drawText( (width() / 2) - (text_rect.width() / 2), (int)(height() * 0.4), 
	       QString::number( _number + 1) );


   p.setPen( QPen( levelColor ) );

   text_rect = p.boundingRect( 0, 0, width(), height(), 0,  
           QString::number( _levelDisplay ) );
  
   p.drawText( (width() / 2) - (text_rect.width() / 2), 
           (int)(height() * 0.8), QString::number( _levelDisplay ) );
   
}

void Channel::mousePressEvent( QMouseEvent * e )
{
  
	if(e->button() != LeftButton) {
		e->ignore();
		return;
	}
  emit selectStart( _number, e->state() & Qt::ControlButton );
  
  QStoredDrag *d = new QStoredDrag( "channel", this );
  d->setEncodedData( QByteArray(1)[0] = 0 );
  d->drag();
}

void Channel::dragEnterEvent(QDragEnterEvent* event)
{
  if( event->provides( "channel" ) )  //use a mime type to check the drag
  {
    emit selectEnter( _number );
    event->accept(true);
  }      
}

void Channel::dropEvent(QDropEvent* event)
{
  if( event->provides( "channel" ) )
  {
    emit selectEnd();
    event->accept(true);
  }
      
}

QSizePolicy Channel::sizePolicy() const
{
    return QSizePolicy( QSizePolicy::Fixed, QSizePolicy::Fixed );
}

QSize Channel::sizeHint() const
{
  return QSize( 50, 50 );
}


